package kr.gmtc.eyegw.controller;

import java.nio.charset.Charset;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.LinkedBlockingQueue;

import javax.annotation.Resource;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.context.event.ApplicationReadyEvent;
import org.springframework.context.ApplicationListener;
import org.springframework.context.event.ContextClosedEvent;
import org.springframework.context.event.EventListener;
import org.springframework.core.annotation.Order;
import org.springframework.stereotype.Component;

import gmt.common.type.LogLevelType;
import gmt.common.util.GmtUtils;
import gmt.common.vo.MsgObjVO;
import gmt.controller.GmtController;
import gmt.io.client.event.GmtClientEventListener;
import gmt.io.server.event.GmtServerEventListener;
import gmt.io.udp.event.GmtUDPEventListener;
import gmt.logger.GmtLogManager;
import kr.gmtc.eyegw.config.ServiceConfig;
import kr.gmtc.eyegw.db.DB1Connection;
import kr.gmtc.eyegw.file.TextReader;
import kr.gmtc.eyegw.file.TextReaderMergeAndSort;
import kr.gmtc.eyegw.frame.CustomThread;
import kr.gmtc.eyegw.frame.StatusReporter;

@Component("controller")
public class MainController extends GmtController implements ApplicationListener<ContextClosedEvent> {

	// 프레임 워크 구성요소 //
	/** 서비스 설정, {@code application.yml} */
	@Resource(name = "ServiceConfig")
	private ServiceConfig serviceConfig;
	
	/* 상태 확인 스테드 */
	private StatusReporter stateReporter;
	
	/* 실생상태 확인 */
	private boolean running = false;
	
	/* UDP test 소켓 */
	//UDPEchoClient UDPSocket ;
	
	private long lSendCnt = 0;
	
	@Resource(name = "packetQ")
	Queue<String> packetQ;
	
	@Value("${asde.filepath1}")
	private String sfilepath1;
	
	@Value("${asde.filepath2}")
	private String sfilepath2;
	
	public Queue<Map<String, String>> fileQ;
	
	// DB //
	/**
	 * DB1 Connection<br>
	 * application.yml에 db.db1.use가 true로 설정되어야 사용가능<br>
	 * db.db1.use가  false일 경우 주석처리 (Bean 런타임 에러 발생)
	 * @see DB1Connection
	 */
	@Resource(name = "DB1Connection")
	private DB1Connection db1Connection;

	
	// 업무처리 //
	/** 업무 처리 스레드 */
	private CustomThread bizThread;

		
	/**
	 * 수정 금지<br>
	 * Spring Bean 객체 생성, logger 초기화
	 */
	public MainController(@Value("${root}") String path) {
		super(path);
		
		GmtLogManager.setpath(path);
	}
	
	/**
	 * 체크섬 검증<br>
	 * 끝에서 3자리 가져와서 검증<br>
	 * 체크섬 검증 옵션이 false이면 무조건 true 반환
	 * @param msg 메시지
	 * @param spliter 체크섬 구분자
	 * @return 체크섬 검증 결과, 체크섬 확인옵션이 false면 무조건 true 리턴
	 */
	private boolean verifyChecksum(final String msg, final String originalMsg) {
		// 메시지 길이 체크
		if (msg == null || msg.trim().isEmpty()) {
			return false;
		}
		// 체크섬 검증
		try {
			char splitChar = msg.charAt(msg.length() - 4);
			if (splitChar == '*') {
				String inputChecksum = msg.substring(msg.length() - 2, msg.length());
				//String realChecksum = GmtUtils.makeCRC(originalMsg.getBytes());
				String realChecksum = GmtUtils.makeCRC(originalMsg);
				return inputChecksum.equals(realChecksum);
			}
		} catch (@SuppressWarnings("unused") Exception e) {
			// Nothing to do
		}
		return false;
	}
	
	/**
	 * 디코드<br>
	 * null safe
	 * @param rawMsg 수신 한 메시지
	 * @return 디코드 된 메시지, 비어있는 메지시는 빈 String 반환
	 */
	private String decodeString(final byte[] rawMsg, Charset charset) {
		// 메시지 길이 체크
		if (rawMsg == null || rawMsg.length < 1) {
			return "";
		}
		// 캐릭터 셋에 맞게 디코드
		return new String(rawMsg, charset).trim();
	}
	
	/**
	 * 실행 (1순위)<br>
	 * 프로그램  초기화
	 */
	@Order(0)
	@EventListener(ApplicationReadyEvent.class)
	public void initialize() {
		try {
			// 상태보고 스레드 생성, 1분간격으로 보고
			stateReporter = new StatusReporter(this, CustomThread.SLEEP_MINIUTE);
			
			// 업무 스레드 생성, 대기 없이 무한반복
			bizThread = new CustomThread("bizThread", this, CustomThread.NO_SLEEP, this::bizWork, null, false);
			
			// UDP소켓 객체 생성
			//UDPSocket = new UDPEchoClient();
					
			fileQ =  new LinkedBlockingQueue<Map<String, String>>();
			
		} catch(Exception e){
			logger.writeLevelLog("[MainController] Prepare Fail " + e.getMessage(), LogLevelType.LOG_ERROR, "AllLog");
			System.exit(1);
		}
	}
	
	
	/**
	 * 실행 (2순위)<br>
	 * 프로그램 실행
	 */
	@Override
	@Order(1)
	@EventListener(ApplicationReadyEvent.class)
	public void start() {
		try {
			running = true;
			super.start();
			
			// 상태보고 스레드 시작
			//stateReporter.start();

			// 업무로직 시작
			bizThread.start();
			
			// UDP수신 시작
			//UDPSocket.start();
			
			getFileBufer(fileQ, sfilepath1, sfilepath2);
			
		} catch (Exception e) {
			logger.writeLevelLog("[MainController] Start Fail " + e.getMessage(), LogLevelType.LOG_ERROR, "AllLog");
			System.exit(1);
		}
	}
	
	/**
	 * 업무 로직<br>
	 * 자동으로 무한반복 하므로 내부에 while문 필요 없음.<br>
	 * 예외처리를 하지 않을 경우 {@link CustomThread} 내부에서 로그 처리
	 */
	private void bizWork() throws InterruptedException {
		
		String msg = "";
		
		// 읽어드린 데이터를 모두 전송한경우 초기화
		if(fileQ.peek() == null) {
			getFileBufer(fileQ, sfilepath1, sfilepath2);
			lSendCnt = 0;
		}
		
		try {
			 
			int iSleep = 0;
			
			// 클라이언트 포트 브로드캐스트
			MsgObjVO msgVO = new MsgObjVO();
			if(fileQ.peek() != null) {
				
				Map<String, String> qMap = new HashMap<String, String>();
				Map<String, String> qMap_after = new HashMap<String, String>();
				
				qMap = fileQ.poll();
				
				//msgVO.setMsg(qMap.get("data").getBytes(serviceConfig.getClientSendCharset()));
				
				String sHexData = qMap.get("data").replaceAll("\\p{Z}", "");
				
				msgVO.setMsg(new java.math.BigInteger(sHexData, 16).toByteArray());
				broadcastUDP(msgVO);
				
				// 다음 전송까지 대기 시간 계산
				qMap_after = fileQ.peek();
				
				String sTime_befre = qMap.get("time");
				String sTime_after = qMap_after.get("time");
				
				Date tb = new SimpleDateFormat("yyyyMMddHHmmss").parse(sTime_befre);
		        Date ta = new SimpleDateFormat("yyyyMMddHHmmss").parse(sTime_after);
				
		        long diffSec = (ta.getTime() - tb.getTime());
		        		
		        logger.writeLevelLog("[MainController-bizWork] send log : " + "***전송:" + sTime_befre + "***다음:" + sTime_after + "***차이:" + diffSec, LogLevelType.LOG_INFO, "AllLog");
		        
//		        if(diffSec <0 ) {
//		        	String sTim = "" ;
//		        }
				lSendCnt ++;
				
//				System.out.print("\n***" + lSendCnt + "건 전송*** " + qMap.get("data"));
				
				Thread.sleep(diffSec);
			}
			
		} catch (InterruptedException e) {
			if (running) {
				logger.writeLevelLog("[MainController-bizWork] InterruptedException : " + e.getMessage(), LogLevelType.LOG_ERROR, "AllLog");
			} else { // 종료 중 인터럽트 발생 시 throw
				System.out.println(e);
			}
		} catch (Exception e) {
			if (running) {
				logger.writeLevelLog("[MainController-bizWork] running process Exception : " + e.getMessage(), LogLevelType.LOG_ERROR, "AllLog");
			} else { // 종료 중 인터럽트 발생 시 throw
				System.out.println(e);
			}
		}
		
		
	}
	
	
	public void getFileBufer(Queue<Map<String, String>> fileQ, String sFilePath1, String sFilePath2) {
		
		//TextReader rf = new TextReader(file1, file2, sFilePath1, sFilePath2);
		TextReaderMergeAndSort rf = new TextReaderMergeAndSort(fileQ, sFilePath1, sFilePath2);
		rf.readFile();
	   // System.out.println(result);

	}
	
	public static byte[] binaryStringToByteArray(String s) {
        int count = s.length() / 8;
        byte[] b = new byte[count];
        for (int i = 1; i < count; ++i) {
            String t = s.substring((i - 1) * 8, i * 8);
            b[i - 1] = binaryStringToByte(t);
        }
        return b;
    }
	
	
	public static byte binaryStringToByte(String s) {
        byte ret = 0, total = 0;
        for (int i = 0; i < 8; ++i) {
            ret = (s.charAt(7 - i) == '1') ? (byte) (1 << i) : 0;
            total = (byte) (ret | total);
        }
        return total;
    }
	

	
	/** 종료 처리 */
	@Override
	public void stop() {
		super.stop();
		
		// UDP소켓 종료
//		try {
//			UDPSocket.start();
//		} catch (IOException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}
		
	}
	
	/** 종료 이벤트 (강제종료시 수행 안됨) */
	@Override
	public void onApplicationEvent(@SuppressWarnings("unused") ContextClosedEvent event) {
		this.stop();
		logger.writeLevelLog("====================== SYSTEM STOPED ======================", LogLevelType.LOG_INFO, "AllLog");
	}
	
	/** 클라이언트 포트 데이터 수신 */
	@Override
	protected void clientDataIn(MsgObjVO vo) {
		if (vo.getMsg().length < 1) {
			return;
		}
		
		try {
			// 메시지 디코드
			final String msg = decodeString(vo.getMsg(), serviceConfig.getClientReceiveCharset());
			// 체크섬 제거한 데이터 추출
			String originalMsg = msg.substring(0, msg.length() - 3);
			// 체크섬 검증
			if (serviceConfig.isClientReceiveChecksum() && !verifyChecksum(msg, originalMsg)) {
				return; // 체크섬 에러 발생 시 무시
			}

			// TODO: 메시지 처리

		} catch (Exception e) {
			logger.writeLevelLog("[clientDataIn] Unknown Exception Occur2. " + e.getMessage(), LogLevelType.LOG_ERROR, "AllLog");
		}
		
	}
	
	/** 서버 포트 데이터 수신 */
	@Override
	protected void serverDataIn(MsgObjVO vo) {
		if (vo.getMsg().length < 1) {
			return;
		}
		
		try {
			// 메시지 디코드
			final String msg = decodeString(vo.getMsg(), serviceConfig.getClientReceiveCharset());
			// 체크섬 제거한 데이터 추출
			String originalMsg = msg.substring(0, msg.length() - 3);
			// 체크섬 검증
			if (serviceConfig.isServerReceiveChecksum() && !verifyChecksum(msg, originalMsg)) {
				return; // 체크섬 에러 발생 시 무시
			}

			// TODO: 메시지 처리

		} catch (Exception e) {
			logger.writeLevelLog("[serverDataIn] Unknown Exception Occur. " + e.getMessage(), LogLevelType.LOG_ERROR, "AllLog");
		}
	}
	
	/** 시리얼 포트 데이터 수신 */
	@Override
	protected void serialDataIn(MsgObjVO vo) {
		if (vo.getMsg().length < 1) {
			return ;
		}
		
		try {
			// TODO: 메시지 처리
		}catch (Exception e) {
			logger.writeLevelLog("[serialDataIn] Unknown Exception Occur. " + e.getMessage(), LogLevelType.LOG_ERROR, "AllLog");
		}
	}
	
	
	/** Client 이벤트 */
	@SuppressWarnings({ "unused", "synthetic-access"})
	@Override
	protected GmtClientEventListener createClientEventListener() {
		return new GmtClientEventListener() {
			@Override
			public void disconnected(String ip, int port, int statusCode, String description) {
				String msg = String.format("[Client Disconnect : IP[%s] / PORT [%d]", ip, port);
				logger.writeLevelLog(msg, LogLevelType.LOG_INFO, "AllLog");
			}
			@Override
			public void connected(String ip, int port, int statusCode, String description) {
				String msg = String.format("[Client Connect : IP[%s] / PORT [%d]", ip, port);
				logger.writeLevelLog(msg, LogLevelType.LOG_INFO, "AllLog");
			}
			@Override
			public void readyToSend() {}
			@Override
			public void error(String ip, int port, int errorCode, String description) {}
			@Override
			public void dataIn() {}
			@Override
			public void connectionStatus(String ip, int port, String connEvent, int statusCode, String description) {}
			@Override
			public void SSLStatus() {}
			@Override
			public void SSLServerAuthentication() {}
		};
	}
	
	/** Server 이벤트 */
	@SuppressWarnings({ "unused", "synthetic-access"})
	@Override
	protected GmtServerEventListener createServerEventListener() {
		return new GmtServerEventListener() {
			@Override
			public void disconnected(int serverPort, String remoteHost, int remotePort) {
				String msg = String.format("[SERVER %d ] SC Disconnected.. IP : %s PORT : %d", serverPort, remoteHost, remotePort);
				logger.writeLevelLog(msg, LogLevelType.LOG_INFO, "AllLog");
			}
			@Override
			public void connected(int serverPort, String remoteHost, int remotePort) {
				String msg = String.format("[SERVER %d ] SC Connected.. IP : %s PORT : %d", serverPort, remoteHost, remotePort);
				logger.writeLevelLog(msg, LogLevelType.LOG_INFO, "AllLog");
			}
			@Override
			public void readyToSend(int port) {}
			@Override
			public void error(int serverPort, String remoteHost, int remotePort, int errorCode, String description) {}
			@Override
			public void dataIn(int serverPort, String remoteHost, int remotePort, String msg) {}
			@Override
			public void SSLClientAuthentication(int serverPort) {}
		};
		
	}

	@Override
	protected void udpDataIn(MsgObjVO vo) {
		if (vo.getMsg().length < 1) {
			return;
		}
		
		try {
			// 메시지 디코드
			final String msg = decodeString(vo.getMsg(), serviceConfig.getClientReceiveCharset());
//			// 체크섬 제거한 데이터 추출
//			String originalMsg = msg.substring(0, msg.length() - 3);
//			// 체크섬 검증
//			if (serviceConfig.isClientReceiveChecksum() && !verifyChecksum(msg, originalMsg)) {
//				return; // 체크섬 에러 발생 시 무시
//			}

			// TODO: 메시지 처리
			packetQ.add(msg);
			
		} catch (Exception e) {
			logger.writeLevelLog("[clientDataIn] Unknown Exception Occur1. " + e.getMessage(), LogLevelType.LOG_ERROR, "AllLog");
		}
	}

	@Override
	protected GmtUDPEventListener createUDPEventListener() {
		GmtUDPEventListener listener = new GmtUDPEventListener() {

			@Override
			public void readyToSend() {
				// TODO Auto-generated method stub
				
			}

			@Override
			public void error(String ip, int port, int errorCode, String description) {
				// TODO Auto-generated method stub
				
			}

			@Override
			public void disconnected(String ip, int port, int statusCode, String description) {
				String msg = "[UDP Socket Disconnect : IP[" + ip + "] / PORT [" + String.valueOf(port) + "] ";
				logger.writeLevelLog(msg, LogLevelType.LOG_INFO, "AllLog");
			}

			@Override
			public void dataIn(int serverPort, String remoteHost, int remotePort, String msg) {
				String msgd = "[UDP Socket d : IP[" + remoteHost + "] / PORT [" + String.valueOf(remotePort) + "] |msg:"+ msg;
				logger.writeLevelLog(msgd, LogLevelType.LOG_INFO, "AllLog");
				
			}

			@Override
			public void connectionStatus(String ip, int port, String connEvent, int statusCode, String description) {
				// TODO Auto-generated method stub
				
			}

			@Override
			public void connected(String ip, int port, int statusCode, String description) {
				String msg = "[UDP Socket Connect : IP[" + ip + "] / PORT [" + String.valueOf(port) + "] ";
				logger.writeLevelLog(msg, LogLevelType.LOG_INFO, "AllLog");
				
			}
			
		};
		
		return listener;
	}
}


